<?php

require_once("binary.php");

$ROMADDR = 0;
$RAMADDR = $ROMADDR+ 0x100000; // size of PoP ROM
$IOADDR = $RAMADDR + 0x20000;
$ENDADDR = $IOADDR + 0x8000;

// [SNES]
// convert linear ROM address to bank+offset
function laddr($ad){
	global $ROMADDR,$RAMADDR,$IOADDR,$ENDADDR;
	if ($ad >= $ROMADDR and $ad < $RAMADDR) {
		// ROM
		$sg=$ad>>15; $of=($ad&0x7FFF)|0x8000;
	} elseif ($ad >= $RAMADDR and $ad < $IOADDR) {
		// RAM
		$ra = $ad-$RAMADDR;
		$sg = ($ra>>16)+0x7E;
		$of = $ra&0xFFFF;
	} elseif ($ad >= $IOADDR and $ad < $ENDADDR) {
		// I/O and expansion
		$sg = 0x00; // Anything in 0x00..0x7D
		$of = $ad-$IOADDR;
	}
	return sprintf("%02X:%04X",$sg,$of);
}

// [SNES]
// Map bank:offset to linear addresses.
// ROM addresses match the ROM image.
// Other addresses are just unique.
function addr_to_linear($sg,$of) {
	global $ROMADDR,$RAMADDR,$IOADDR,$ENDADDR;
//	printf("addr_to_linear: seg=0x%02X, off=0x%04X\r\n",$sg,$of);
	if($sg>=0 and $sg<0x7E) {
		if ($of>=0x8000 and $of<0x10000) {
			// ROM
			return $ROMADDR + ($sg<<15) + ($of&0x7FFF);
		}
		if ($of>=0x0000 and $of<0x2000) {
			// Mirror of first 8k RAM
			return $RAMADDR + $of;
		}
		if ($of>=0x2000 and $of<0x8000) {
			// I/O and expansion
			return $IOADDR + $of;
		}
	}
	if($sg>=0x7E and $sg<0x80){
		// RAM
		return $RAMADDR + (($sg-0x7E)<<16) + $of;
	}
	printf("ERROR in addr_to_linear! Can't map %02X:%04X\r\n",$sg,$of);
	return -1;
}

// [65C816]
class flags{
/*
A class to store bytes that may have unknown bits.
bit is 0 -> 0 in min and max
bit is 1 -> 1 in min and max
bit is unknown -> 0 in min, 1 in max
*/
	var $min,$max;
	function flags($min_=0,$max_=0xFF){ $this->min=$min_; $this->max=$max_; }
	function rep($v){ $this->min&=~$v; $this->max&=~$v; }
	function sep($v){ $this->min|=$v; $this->max|=$v; }
	function get($v){ if(($this->min&$v)!=($this->max&$v)) return null; else return $this->min&$v; }
	function undef(){ $this->min=0; $this->max=0xFF; }
};

// Filenames:
$fn=$argv[1] or die;
$fnannot=$argv[2] or die;

// Load ROM image:
$f=file_get_contents($fn);
$si=strlen($f);
// Remove header if present:
switch($si&0x7FFF){
	case 0: break;
	case 0x200: $f=substr($f,0x200); break;
	default: die("ROM has wrong size!");
}

// Load annotations:
$fannot=file_get_contents($fnannot);
$fannot=explode("\r\n",$fannot);
$hints=array();
$dec=array();
$labels=array();
$threads=array();
foreach($fannot as $l){
	@list($ad,$op,$arg,$arg2)=explode("\t",$l);
	sscanf($ad,"%02X:%04X",$sg,$of);
	//if(!($of&0x8000)) continue;
	//$ad=$sg<<15|$of&0x7FFF;
	$ad=addr_to_linear($sg,$of);
	switch($op){
		case "HINT": $hints[$ad][]=$arg; break;
		case "DATA":
			sscanf($arg,"%02X:%04X",$sg,$of);
			//if(!($of&0x8000)) continue;
			//$ad2=$sg<<15|$of&0x7FFF;
			$ad2=addr_to_linear($sg,$of);
			switch($arg2){
				case "STRING":
					$dt=substr($f,$ad,$ad2-$ad); $next[$ad]=$ad2;
					$dec[$ad]="STRING \"$dt\"";
				break;
				case "BYTE":
					$next[$ad]=$ad2;
					$line="BYTE ";
					for($i=$ad,$sep="";$i<$ad2;$i++,$sep=",") $line.=$sep."$".h2(ord($f[$i]),2);
					$dec[$ad]=$line;
				break;
				case "WORD":
					$next[$ad]=$ad2;
					$line="WORD ";
					for($i=$ad,$sep="";$i<$ad2;$i+=2,$sep=",") $line.=$sep."$".h2(ord($f[$i+1]),2).h2(ord($f[$i]),2);
					$dec[$ad]=$line;
				break;
			}
		break;
		case "LABEL":
			$labels[$ad][]=$arg;
		break;
		case "CODE":
			$threads[]=array($ad,new flags(),-1);
		break;
		case "SKIP":
			$skips[$ad]=(int)$arg;
		break;
	}
}
//print_r($hints);

// [65C816]
// Instructions
// nn: 1-byte constant
// nnnn: 2-byte constant
// nnnnnn: 3-byte constant: bank number + 2-byte address
// a: 1 or 2-byte constant, depending on accumulator size
// x: 1 or 2-byte constant, depending on index register size
// ee: 1-byte PC-relative address
$ops=array(
//0x00=>'BRK $nn',
//0x01=>'ORA ($nn,X)',
//0x02=>'COP $nn',
0x03=>'ORA $nn,S',
0x04=>'TSB $nn',
0x05=>'ORA $nn',
0x06=>'ASL $nn',
//0x07=>'ORA [$nn]',
0x08=>'PHP',
0x09=>'ORA #$a',
0x0A=>'ASL',
//0x0B=>'PHD',
0x0C=>'TSB $nnnn',
0x0D=>'ORA $nnnn',
0x0E=>'ASL $nnnn',
0x0F=>'ORA $nnnnnn',
0x10=>'BPL $ee',
//0x12=>'ORA ($nn)',
//0x13=>'ORA ($nn,S),Y',
//0x14=>'TRB $nn',
0x15=>'ORA $nn,X',
//0x16=>'LDA $nnnn,X',
0x17=>'ORA [$nn],Y',
0x18=>'CLC',
0x19=>'ORA $nnnn,Y',
0x1A=>'INC',
0x1B=>'TCS',
0x1C=>'TRB $nnnn',
0x1D=>'ORA $nnnn,X',
0x1F=>'ORA $nnnnnn,X',
0x20=>'JSR $nnnn',
//0x21=>'AND ($nn,X)',
0x22=>'JSR $nnnnnn',
0x23=>'AND $nn,S',
0x25=>'AND $nn',
0x26=>'ROL $nn',
0x27=>'AND [$nn]',
0x28=>'PLP',
0x29=>'AND #$a',
0x2A=>'ROL',
0x2C=>'BIT $nnnn',
0x2D=>'AND $nnnn',
0x2E=>'STA $nnnn,X',
0x30=>'BMI $ee',
//0x31=>'AND ($nn),Y',
0x35=>'AND $nn,X',
0x38=>'SEC',
0x39=>'AND $nnnn,Y',
0x3A=>'DEC',
0x3B=>'TSC',
0x3D=>'AND $nnnn,X',
0x3F=>'AND $nnnnnn,X',
0x40=>'RTI',
0x43=>'EOR $nn,S',
0x46=>'LSR $nn',
0x48=>'PHA',
0x49=>'EOR #$a',
0x4A=>'LSR',
0x4B=>'PHK',
0x4C=>'JMP $nnnn',
0x4D=>'EOR $nnnn',
0x4E=>'LSR $nnnn',
//0x50=>'BVC $ee',
//0x53=>'EOR ($nn,S),Y',
//0x54=>'MVN $nn,$nn',
0x55=>'EOR $nn,X',
//0x56=>'LSR $nn,X',
//0x58=>'CLI',
0x5A=>'PHY',
0x5B=>'TCD',
0x5C=>'JMP $nnnnnn',
0x5D=>'EOR $nnnn,X',
0x60=>'RTS',
0x63=>'ADC $nn,S',
0x64=>'STZ $nn',
0x65=>'ADC $nn',
0x68=>'PLA',
0x69=>'ADC #$a',
0x6A=>'ROR',
0x6B=>'RTL',
0x6D=>'ADC $nnnn',
0x6F=>'ADC $nnnnnn',
0x74=>'STZ $nn,X',
0x75=>'ADC $nn,X',
0x79=>'ADC $nnnn,Y',
0x7A=>'PLY',
0x7C=>'JMP ($nnnn,X)',
0x7D=>'ADC $nnnn,X',
0x7F=>'ADC $nnnnnn,X',
0x80=>'BRA $ee',
0x83=>'STA $nn,S',
0x84=>'STY $nn',
0x85=>'STA $nn',
0x86=>'STX $nn',
0x87=>'STA [$nn]',
0x88=>'DEY',
0x8A=>'TXA',
0x8B=>'PHB',
0x8C=>'STY $nnnn',
0x8D=>'STA $nnnn',
0x8E=>'STX $nnnn',
0x8F=>'STA $nnnnnn',
0x90=>'BCC $ee',
//0x91=>'STA ($nn),Y',
0x93=>'STA ($nn,S),Y',
0x94=>'STY $nn,X',
0x95=>'STA $nn,X',
0x97=>'STA [$nn],Y',
0x98=>'TYA',
0x99=>'STA $nnnn,Y',
0x9A=>'TXS',
0x9B=>'TXY',
0x9C=>'STZ $nnnn',
0x9D=>'STA $nnnn,X',
0x9E=>'STZ $nnnn,X',
0x9F=>'STA $nnnnnn,X',
0xA0=>'LDY #$x',
//0xA1=>'ASL $nnnn',
0xA2=>'LDX #$x',
0xA3=>'LDA $nn,S',
0xA4=>'LDY $nn',
0xA5=>'LDA $nn',
0xA6=>'LDX $nn',
0xA7=>'LDA [$nn]',
0xA8=>'TAY',
0xA9=>'LDA #$a',
0xAA=>'TAX',
0xAB=>'PLB',
0xAC=>'LDY $nnnn',
0xAD=>'LDA $nnnn',
0xAE=>'LDX $nnnn',
0xAF=>'LDA $nnnnnn',
0xB0=>'BCS $ee',
0xB1=>'LDA ($nn),Y',
0xB2=>'LDA ($nn)',
0xB3=>'LDA ($nn,S),Y',
0xB4=>'LDY $nn,X',
0xB5=>'LDA $nn,X',
0xB7=>'LDA [$nn],Y',
0xB9=>'LDA $nnnn,Y',
0xBA=>'TSX',
0xBB=>'TYX',
0xBC=>'LDY $nnnn,X',
0xBD=>'LDA $nnnn,X',
0xBE=>'LDX $nnnn,Y',
0xBF=>'LDA $nnnnnn,X',
0xC0=>'CPY #$x',
0xC2=>'REP #$nn',
0xC3=>'CMP $nn,S',
0xC5=>'CMP $nn',
0xC6=>'DEC $nn',
0xC7=>'CMP [$nn]',
0xC8=>'INY',
0xC9=>'CMP #$a',
0xCA=>'DEX',
0xCD=>'CMP $nnnn',
0xCE=>'DEC $nnnn',
0xCF=>'CMP $nnnnnn',
0xD0=>'BNE $ee',
0xD1=>'CMP ($nn),Y',
0xD9=>'CMP $nnnn,Y',
0xDA=>'PHX',
0xDD=>'CMP $nnnn,X',
0xDE=>'DEC $nnnn,X',
0xDF=>'CMP $nnnnnn,X',
0xE0=>'CPX #$x',
0xE2=>'SEP #$nn',
0xE3=>'SBC $nn,S',
0xE5=>'SBC $nn',
0xE6=>'INC $nn',
0xE8=>'INX',
0xE9=>'SBC #$a',
0xEB=>'XBA',
0xEC=>'CPX $nnnn',
0xED=>'SBC $nnnn',
0xEE=>'INC $nnnn',
0xF0=>'BEQ $ee',
//0xF2=>'SBC ($nn)',
0xF4=>'PEA $nnnn',
0xF5=>'SBC $nn,X',
0xF9=>'SBC $nnnn,Y',
0xFA=>'PLX',
0xFB=>'XCE',
0xFD=>'SBC $nnnn,X',
0xFF=>'SBC $nnnnnn,X',
);
// instructions that unconditionally jump/return
$opend=array(0x40=>true,0x4C=>true,0x60=>true,0x6B=>true,0x7C=>true,0x80=>true,0x5C=>true);
// instructions that jump to a known address
$opjum=array(0x10=>1,0x20=>2,0x22=>3,0x30=>1,0x4C=>2,0xD0=>1,0x80=>1,0x90=>1,0xB0=>1,0xF0=>1,0x50=>1);

// [65C816]
// Hardware vectors
$vecs=array(
	"IRQ_nat"=>0xFFEE,
	"NMI_nat"=>0xFFEA,
	"ABORT_nat"=>0xFFE8,
	"BRK_nat"=>0xFFE6,
	"COP_nat"=>0xFFE4,
	"IRQ_BRK_emu"=>0xFFFE,
	"RESET"=>0xFFFC,
	"NMI_emu"=>0xFFFA,
	"ABORT_emu"=>0xFFF8,
	"COP_emu"=>0xFFF4,
);

$databank=-1;
foreach($vecs as $k=>$v){
	//$p=$v&0x7FFF;
	$p=addr_to_linear(0x00,$v);
	$of=getint2($f,$p);
	$ad=addr_to_linear(0x00,$of);
	//if($of&0x8000){
	if($ad>=$ROMADDR and $ad<$RAMADDR){
		$threads[]=array($ad,new flags(),$databank);
		$labels[$ad][]=$k;
	}else{
		printf("Warning, vector %s points to RAM: 0x%04X\r\n",$v);
	}
}

// Main disassembly loop.
$allend=false;
$end=false;
list($p,$flags)=array_pop($threads);
//$flags=new flags();
while(!$allend){
	$sa=$p;
	// [65C816] Hints
	if(isset($hints[$sa])){
//print "HINTS!".h2($p,8);
		foreach($hints[$sa] as $hint){
			switch($hint){
				case "A=8": $flags->sep(0x20); break;
				case "X=8": $flags->sep(0x10); break;
				case "A=16": $flags->rep(0x20); break;
				case "X=16": $flags->rep(0x10); break;
			}
			if(substr($hint,0,2)==="B="){
				$databank = hexdec(substr($hint,2));
			}
		}
	}
	// [65C816] Opcodes
	$op=getint1($f,&$p);
	if(isset($ops[$op])){
		$line=$ops[$op];
		if(isset($opend[$op])) $end=true;
		$v=null;
		$opsize=null;
		if(strpos($line,"a")!==false){
			$f20=$flags->get(0x20);
			if($f20===null){ $end=true; $line.=" !unknown flags!"; }else
			$line=str_replace("a",$f20?"nn":"nnnn",$line);
		}
		if(strpos($line,"x")!==false){
			$f10=$flags->get(0x10);
			if($f10===null){ $end=true; $line.=" !unknown flags!"; }else
			$line=str_replace("x",$f10?"nn":"nnnn",$line);
		}
		if(strpos($line,"ee")!==false){
			$v=getint1($f,$p); if($v&0x80) $v-=0x100;
			$v=($v+($p&0x7FFF|0x8000))&0xFFFF;
			$line=str_replace("ee",h2($v,4),$line);
		}elseif(strpos($line,"nnnnnn")!==false){
			$of=getint2($f,$p); $sg=getint1($f,$p);
			$line=str_replace("nnnnnn",h2($sg,2).":".h2($of,4),$line);
			$v=$sg<<16|$of;
			$opsize=3;
		}elseif(strpos($line,"nnnn")!==false){
			$v=getint2($f,$p);
			$line=str_replace("nnnn",h2($v,4),$line);
			$opsize=2;
		}elseif(strpos($line,'nn,$nn')!==false){
			$v1=getint1($f,$p);
			$v2=getint1($f,$p);
			$line=str_replace('nn,$nn',h2($v1,2).",$".h2($v2,2),$line);
		}elseif(strpos($line,"nn")!==false){
			$v=getint1($f,$p);
			$line=str_replace("nn",h2($v,2),$line);
			$opsize=1;
		}
		// [65C816] REP/SEP
		if($op==0xC2) $flags->rep($v);
		if($op==0xE2) $flags->sep($v);
		$ref_ad = null;
		$js = 0;
		if(isset($opjum[$op])){
			// [65C816] jumps
			$js=$opjum[$op];
			switch($js){
				case 1: $sg=$p>>15; $of=$v; break;
				case 2: $sg=$p>>15; $of=$v; break;
				case 3: $sg=$v>>16; $of=$v&0xFFFF; break;
				default: die("Bad jump size: ".$js);
			}
			$ref_ad = addr_to_linear($sg,$of);
		} elseif (strpos($line,"#")===false and $v!==null) {
			// data references
			switch($opsize) {
				case 1: $sg=0; $of=$v; break; // zero page is always in bank 0
				case 2: $sg=$databank; $of=$v; break;
				case 3: $sg=$v>>16; $of=$v&0xFFFF; break;
				default: die("Bad operand size: ".$js);
			}
			$ref_ad = addr_to_linear($sg,$of);
		}
		if ($ref_ad !== null) {
			if ($ref_ad>=0) {
				if($of>=$ROMADDR and $of<$RAMADDR){
					if ($js) { // was a jump
						$threads[]=array($ref_ad,$flags,$databank);
					}
					if(isset($labels[$ref_ad])){
						$line.=" ; ".implode(" ; ",$labels[$ref_ad]);
					}
				}else{
					printf("Warning, jump into RAM: from %s to %s\r\n",laddr($p),laddr($ref_ad));
				}
			} else {
				$line .= " ; unknown B, can't map";
			}
		}
		// [65C816] instructions that leave flags in undefined state: JSR, PLP
		if($op==0x28 /*or $op==0x20 or $op==0x22 or $op==0xFC*/) $flags->undef();
		if($op===0xAB){ // PLB
			if (!isset($hints[$sa])) {
				// TODO: Figure out the simple cases...
				$databank=-1;
				$line .= sprintf(" ; B=???");
			} else {
				$line .= sprintf(" ; B=0x%02X",$databank);
			}
		}
	}else{
		$line="Unknown opcode: ".h2($op,2); $end=true;
	}

	if($end) $line.="\r\n";
	$hex=str2hex(substr($f,$sa,$p-$sa));
	while(strlen($hex)<12) $hex.=" ";
	$dec[$sa]=$hex.$line;
	$next[$sa]=$p;
		if($op===0x20 or $op===0x22){ //JSR
			// Subroutines that take data from after the JSR:
			if(isset($skips[$ref_ad])){
				$end=true; array_push($threads,array($p+$skips[$ref_ad],$flags,$databank));
			}
		}
	if($end){
		$ok=false;
		while(count($threads)){
//print_r($threads);
			list($p,$flags,$databank)=array_pop($threads);
			if(!isset($dec[$p])){ $ok=true; break; }
		}
		if(!$ok and !count($threads)){
			$allend=true;
		}
		$end=false;
	}
}

ksort($dec);
reset($dec);
$en=0;
//print_r($labels);
while(list($k,$v)=each($dec)){
	//if($k==$start) print "START\r\n";
	if($en<$k) print laddr($en).": ".urlencode(substr($f,$en,$k-$en))." [unexplored]\r\n";
	elseif($en>$k) print laddr($en).": NEGATIVE JUMP!\r\n";
	if(isset($labels[$k])) print implode(": ",$labels[$k]).":\r\n";
	print laddr($k).": ".$v."\r\n";
	$en=$next[$k];
}
